/*
 *  m_monitor.c: Implements tcm/Sentinel-like user searching functionality
 *
 *  Copyright 2003-2004 by W. Campbell
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are
 *  met:
 *
 *  1.Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer. 
 *  2.Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution. 
 *  3.The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission. 
 *
 *  THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 *  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 *  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 *  IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 *  $Id$
 *
 * Revision 1:  Commands ULIST and TESTMASK
 * Revision 2:  Command NLIST
 * Revision 3:  TESTMASK split off into its own module
 *              Merge the regex and match versions into a single module,
 *                with the options handled by getopt().
 *              Command GLIST
 * Revision 4:  Compliance with OPERSPY, Removed Hybrid support, added
 *                Ratbox 2 support.
 */

/* List of ircd includes from ../include/ */
#include "stdinc.h"
#include "client.h"
#include "common.h"     /* FALSE bleah */
#include "ircd.h"
#include "irc_string.h"
#include "numeric.h"
#include "s_conf.h"
#include "s_log.h"
#include "s_serv.h"
#include "send.h"
#include "msg.h"
#include "parse.h"
#include "modules.h"
#include <regex.h>

#ifdef MAPI_RATBOX

#include "s_newconf.h"

#else

#include "handlers.h"

#define report_operspy(x, y, z)		log_operspy((x), (y), (z))
#endif

static int mo_ulist(struct Client *client_p, struct Client *source_p,
                     int parc, char *parv[]);
static int mo_nlist(struct Client *client_p, struct Client *source_p,
                     int parc, char *parv[]);
static int mo_glist(struct Client *client_p, struct Client *source_p,
                     int parc, char *parv[]);

static void ul_usage(struct Client *source_p);
static void nl_usage(struct Client *source_p);
static void gl_usage(struct Client *source_p);

/* ULIST handlers */
static void ul_local_match(struct Client *client_p, struct Client *source_p,
                           char *pattern, int lim);
static void ul_local_regex(struct Client *client_p, struct Client *source_p,
                           char *pattern, int lim, unsigned int flags);
static void ul_global_match(struct Client *client_p, struct Client *source_p,
                            char *pattern, int lim);
static void ul_global_regex(struct Client *client_p, struct Client *source_p,
                            char *pattern, int lim, unsigned int flags);

/* NLIST handlers */
static void nl_local_match(struct Client *client_p, struct Client *source_p,
                           char *pattern, int lim);
static void nl_local_regex(struct Client *client_p, struct Client *source_p,
                           char *pattern, int lim, unsigned int flags);
static void nl_global_match(struct Client *client_p, struct Client *source_p,
                            char *pattern, int lim);
static void nl_global_regex(struct Client *client_p, struct Client *source_p,
                            char *pattern, int lim, unsigned int flags);

/* GLIST handlers */
static void gl_local_match(struct Client *client_p, struct Client *source_p,
                           char *pattern, int lim);
static void gl_local_regex(struct Client *client_p, struct Client *source_p,
                           char *pattern, int lim, unsigned int flags);
static void gl_global_match(struct Client *client_p, struct Client *source_p,
                            char *pattern, int lim);
static void gl_global_regex(struct Client *client_p, struct Client *source_p,
                            char *pattern, int lim, unsigned int flags);

#ifdef MAPI_RATBOX
struct Message ulist_msgtab =
{
  "ULIST", 0, 0, 0, MFLG_SLOW,
  { mg_unreg, mg_not_oper, mg_ignore, mg_ignore, mg_ignore, { mo_ulist, 2 } }
};

struct Message nlist_msgtab =
{
  "NLIST", 0, 0, 0, MFLG_SLOW,
  { mg_unreg, mg_not_oper, mg_ignore, mg_ignore, mg_ignore, { mo_nlist, 2 } }
};

struct Message glist_msgtab =
{
  "GLIST", 0, 0, 0, MFLG_SLOW,
  { mg_unreg, mg_not_oper, mg_ignore, mg_ignore, mg_ignore, { mo_glist, 2 } }
};

mapi_clist_av1 xlist_clist[] =
{
  &ulist_msgtab,
  &nlist_msgtab,
  &glist_msgtab,
  NULL
};

DECLARE_MODULE_AV1(xlist, NULL, NULL, xlist_clist, NULL, NULL, "Hwy v.4 ($Revision$)");

#else /* Ratbox 1.x */

struct Message ulist_msgtab = {
  "ULIST", 0, 0, 2, 0, MFLG_SLOW, 0,
  {m_unregistered, m_not_oper, m_ignore, mo_ulist}
};

struct Message nlist_msgtab = {
  "NLIST", 0, 0, 2, 0, MFLG_SLOW, 0,
  {m_unregistered, m_not_oper, m_ignore, mo_nlist}
};

struct Message glist_msgtab = {
  "GLIST", 0, 0, 2, 0, MFLG_SLOW, 0,
  {m_unregistered, m_not_oper, m_ignore, mo_glist}
};

#ifndef STATIC_MODULES
void
_modinit(void)
{
  mod_add_cmd(&ulist_msgtab);
  mod_add_cmd(&nlist_msgtab);
  mod_add_cmd(&glist_msgtab);
}

void
_moddeinit(void)
{
  mod_del_cmd(&ulist_msgtab);
  mod_del_cmd(&nlist_msgtab);
  mod_del_cmd(&glist_msgtab);
}

const char *_version = "Hwy v.4 ($Revision$)";
#endif /* STATIC_MODULES */

#endif /* MAPI_RATBOX */

#define UL_GLOBAL	0x00000001
#define UL_LOCAL	0x00000002
#define UL_MATCH	0x00000004
#define UL_REGEX	0x00000008
#define UL_INSEN	0x00000010
#define UL_SENS		0x00000020

/*
** mo_ulist
** Syntax:  ULIST [-l|-g] [-m|-r] [-i|-s] [-x limit] user@host
** Defaults:  local scope, limit to 0, match() variant, case insensitive (re)
*/
static int mo_ulist(struct Client *client_p, struct Client *source_p,
                    int parc, char *parv[])
{		 
  int lim = 0;
  char *pattern;
  unsigned int flags = UL_LOCAL|UL_MATCH|UL_INSEN;
  char c;

  opterr = 0; /* Disable error reporting (since it goes to STDERR) */
  optind = 1; /* Always restore the index to the start of the parameters */
  while ( (c = getopt(parc, parv, "lgmrisx:")) != -1)
  {
    switch (c)
    {
      case 'l':
        flags &= ~UL_GLOBAL;
        flags |= UL_LOCAL;
        break;
      case 'g':
        flags &= ~UL_LOCAL;
        flags |= UL_GLOBAL;
        break;
      case 'm':
        flags &= ~UL_REGEX;
        flags |= UL_MATCH;
        break;
      case 'r':
        flags &= ~UL_MATCH;
        flags |= UL_REGEX;
        break;
      case 'i':
        flags &= ~UL_SENS;
        flags |= UL_INSEN;
        break;
      case 's':
        flags &= ~UL_INSEN;
        flags |= UL_SENS;
        break;
      case 'x':
        lim = atoi(optarg);
        break;
      case '?': /* Returned on a syntax error from getopt() */
      default:
        ul_usage(source_p);
        return 0;
        break;
    }
  }

  /* This extracts the pointer to the pattern, which is placed after all
  ** all of the arguments to the ULIST command.
  */
  parc -= optind;
  parv += optind;

  if (parc < 1)
  {
    ul_usage(source_p);
    return 0;
  }

  pattern = parv[0];

  /* Send the information to the proper handlers */
  if (flags & UL_LOCAL)
  {
    if (flags & UL_MATCH)
    {
      ul_local_match(client_p, source_p, pattern, lim);
    }
    else /* UL_REGEX */
    {
      ul_local_regex(client_p, source_p, pattern, lim, flags);
    }
  }
  else /* UL_GLOBAL */
  {
    if (flags & UL_MATCH)
    {
      ul_global_match(client_p, source_p, pattern, lim);
    }
    else /* UL_REGEX */
    {
      ul_global_regex(client_p, source_p, pattern, lim, flags);
    }
  }

  return 0;
}

static void ul_local_match(struct Client *client_p, struct Client *source_p,
                           char *pattern, int lim)
{
  dlink_node *dl;
  struct Client *target_p;
  char *hostname;
  int count = 0;

  hostname = strchr(pattern, '@');
  if (hostname == NULL)
  {
    sendto_one(source_p, ":%s NOTICE %s :Error in /ULIST, please use an '@' "
               "character in the pattern", me.name, source_p->name);
    return;
  }
  *hostname = '\0';
  hostname++;

  DLINK_FOREACH(dl, lclient_list.head)
  {
    target_p = dl->data;

    if ((lim > 0) && (count == lim))
    {
      sendto_one(source_p, ":%s NOTICE %s :ULIST limit reached, %d clients "
                 "shown", me.name, source_p->name, count);
      return;
    }
    if (match(pattern, target_p->username) && match(hostname, target_p->host))
    {
      count++;
      sendto_one(source_p, ":%s NOTICE %s :User %d: %s (%s@%s) [%s]",
                 me.name, source_p->name, count, target_p->name,
                 target_p->username, target_p->host, target_p->user->server);
    }
  }
  sendto_one(source_p, ":%s NOTICE %s :End of ULIST.", me.name, source_p->name);
}

static void ul_local_regex(struct Client *client_p, struct Client *source_p,
                           char *pattern, int lim, unsigned int flags)
{
  int retval;
  dlink_node *dl;
  struct Client *target_p;
  regex_t reg;
  int count = 0;
  char userhost[USERLEN + HOSTLEN + 3];

  if (flags & UL_INSEN)
  {
    retval = regcomp(&reg, pattern, REG_EXTENDED|REG_ICASE|REG_NOSUB);
  }
  else
  {
    retval = regcomp(&reg, pattern, REG_EXTENDED|REG_NOSUB);
  }

  if (retval != 0)
  {
    char errbuf[BUFSIZE];
    regerror(retval, &reg, errbuf, BUFSIZE);
    sendto_one(source_p, ":%s NOTICE %s :ULIST regular expression error: %s",
               me.name, source_p->name, errbuf);
    regfree(&reg);
    return;
  }

  DLINK_FOREACH(dl, lclient_list.head)
  {
    target_p = dl->data;

    if ((lim > 0) && (count == lim))
    {
      sendto_one(source_p, ":%s NOTICE %s :ULIST limit reached, %d clients "
                 "shown", me.name, source_p->name, count);
      return;
    }
    snprintf(userhost, sizeof(userhost), "%s@%s", target_p->username,
             target_p->host);
    /* Using no flags here */
    if (regexec(&reg, userhost, 0, NULL, 0) == 0)
    {
      count++;
      sendto_one(source_p, ":%s NOTICE %s :User %d: %s (%s@%s) [%s]",
                 me.name, source_p->name, count, target_p->name,
                 target_p->username, target_p->host, target_p->user->server);
    }
  }
  sendto_one(source_p, ":%s NOTICE %s :End of ULIST.", me.name, source_p->name);
  regfree(&reg);
}

static void ul_global_match(struct Client *client_p, struct Client *source_p,
                            char *pattern, int lim)
{
  dlink_node *dl;
  struct Client *target_p;
  char *hostname;
  int count = 0;
  char osbuffer[BUFSIZE];

  hostname = strchr(pattern, '@');
  if (hostname == NULL)
  {
    sendto_one(source_p, ":%s NOTICE %s :Error in /ULIST, please use an '@' "
               "character in the pattern", me.name, source_p->name);
    return;
  }
  *hostname = '\0';
  hostname++;

  snprintf(osbuffer, sizeof(osbuffer), "[GLOBAL][MATCH] %s@%s", pattern,
           hostname);
  report_operspy(source_p, "ULIST", osbuffer);

  DLINK_FOREACH(dl, global_client_list.head)
  { 
    target_p = dl->data;

    if (!IsPerson(target_p))
      continue;

    if ((lim > 0) && (count == lim))
    {
      sendto_one(source_p, ":%s NOTICE %s :ULIST limit reached, %d clients "
                 "shown", me.name, source_p->name, count);
      return;
    }
    if (match(pattern, target_p->username) && match(hostname, target_p->host))
    {
      count++;
      sendto_one(source_p, ":%s NOTICE %s :User %d: %s (%s@%s) [%s]",
                 me.name, source_p->name, count, target_p->name,
                 target_p->username, target_p->host, target_p->user->server);
    }
  }
  sendto_one(source_p, ":%s NOTICE %s :End of ULIST.", me.name, source_p->name);
}

static void ul_global_regex(struct Client *client_p, struct Client *source_p,
                            char *pattern, int lim, unsigned int flags)
{
  int retval;
  dlink_node *dl;
  struct Client *target_p;
  regex_t reg;
  int count = 0;
  char userhost[USERLEN + HOSTLEN + 3];
  char osbuffer[BUFSIZE];

  if (flags & UL_INSEN)
  {
    retval = regcomp(&reg, pattern, REG_EXTENDED|REG_ICASE|REG_NOSUB);
  }
  else
  {
    retval = regcomp(&reg, pattern, REG_EXTENDED|REG_NOSUB);
  }

  if (retval != 0)
  {
    char errbuf[BUFSIZE];
    regerror(retval, &reg, errbuf, BUFSIZE);
    sendto_one(source_p, ":%s NOTICE %s :ULIST regular expression error: %s",
               me.name, source_p->name, errbuf);
    regfree(&reg);
    return;
  }

  snprintf(osbuffer, sizeof(osbuffer), "[GLOBAL][REGEX]%s %s",
           (flags & UL_INSEN) ? "[INSENS]" : "[SENS]", pattern);
  report_operspy(source_p, "ULIST", osbuffer);

  DLINK_FOREACH(dl, global_client_list.head)
  {
    target_p = dl->data;

    if (!IsPerson(target_p))
      continue;

    if ((lim > 0) && (count == lim))
    {
      sendto_one(source_p, ":%s NOTICE %s :ULIST limit reached, %d clients "
                 "shown", me.name, source_p->name, count);
      return;
    }
    snprintf(userhost, sizeof(userhost), "%s@%s", target_p->username,
             target_p->host);
    /* Using no flags here */
    if (regexec(&reg, userhost, 0, NULL, 0) == 0)
    {
      count++;
      sendto_one(source_p, ":%s NOTICE %s :User %d: %s (%s@%s) [%s]",
                 me.name, source_p->name, count, target_p->name,
                 target_p->username, target_p->host, target_p->user->server);
    }
  }
  sendto_one(source_p, ":%s NOTICE %s :End of ULIST.", me.name, source_p->name);
  regfree(&reg);
}

static void ul_usage(struct Client *source_p)
{
  sendto_one(source_p, ":%s NOTICE %s :Syntax:  /ULIST [-l|-g] [-m|-r] [-i|-s] "
                       "[-x limit] pattern", me.name, source_p->name);
}

/*
** mo_nlist
** Syntax:  NLIST [-l|-g] [-m|-r] [-i|-s] [-x limit] user@host
** Defaults:  local scope, limit to 0, match() variant, case sensitive (re)
*/
static int mo_nlist(struct Client *client_p, struct Client *source_p,
                    int parc, char *parv[])
{		 
  int lim = 0;
  char *pattern;
  unsigned int flags = UL_LOCAL|UL_MATCH|UL_SENS;
  char c;

  opterr = 0; /* Disable error reporting (since it goes to STDERR) */
  optind = 1; /* Always restore the index to the start of the parameters */
  while ( (c = getopt(parc, parv, "lgmrisx:")) != -1)
  {
    switch (c)
    {
      case 'l':
        flags &= ~UL_GLOBAL;
        flags |= UL_LOCAL;
        break;
      case 'g':
        flags &= ~UL_LOCAL;
        flags |= UL_GLOBAL;
        break;
      case 'm':
        flags &= ~UL_REGEX;
        flags |= UL_MATCH;
        break;
      case 'r':
        flags &= ~UL_MATCH;
        flags |= UL_REGEX;
        break;
      case 'i':
        flags &= ~UL_SENS;
        flags |= UL_INSEN;
        break;
      case 's':
        flags &= ~UL_INSEN;
        flags |= UL_SENS;
        break;
      case 'x':
        lim = atoi(optarg);
        break;
      case '?': /* Returned on a syntax error from getopt() */
      default:
        nl_usage(source_p);
        return 0;
        break;
    }
  }

  /* This extracts the pointer to the pattern, which is placed after all
  ** all of the arguments to the ULIST command.
  */
  parc -= optind;
  parv += optind;

  if (parc < 1)
  {
    nl_usage(source_p);
    return 0;
  }

  pattern = parv[0];

  /* Send the information to the proper handlers */
  if (flags & UL_LOCAL)
  {
    if (flags & UL_MATCH)
    {
      nl_local_match(client_p, source_p, pattern, lim);
    }
    else /* UL_REGEX */
    {
      nl_local_regex(client_p, source_p, pattern, lim, flags);
    }
  }
  else /* UL_GLOBAL */
  {
    if (flags & UL_MATCH)
    {
      nl_global_match(client_p, source_p, pattern, lim);
    }
    else /* UL_REGEX */
    {
      nl_global_regex(client_p, source_p, pattern, lim, flags);
    }
  }

  return 0;
}

static void nl_local_match(struct Client *client_p, struct Client *source_p,
                           char *pattern, int lim)
{
  dlink_node *dl;
  struct Client *target_p;
  int count = 0;

  DLINK_FOREACH(dl, lclient_list.head)
  {
    target_p = dl->data;

    if ((lim > 0) && (count == lim))
    {
      sendto_one(source_p, ":%s NOTICE %s :NLIST limit reached, %d clients "
                 "shown", me.name, source_p->name, count);
      return;
    }
    if (match(pattern, target_p->name))
    {
      count++;
      sendto_one(source_p, ":%s NOTICE %s :User %d: %s (%s@%s) [%s]",
                 me.name, source_p->name, count, target_p->name,
                 target_p->username, target_p->host, target_p->user->server);
    }
  }
  sendto_one(source_p, ":%s NOTICE %s :End of NLIST.", me.name, source_p->name);
}

static void nl_local_regex(struct Client *client_p, struct Client *source_p,
                           char *pattern, int lim, unsigned int flags)
{
  int retval;
  dlink_node *dl;
  struct Client *target_p;
  regex_t reg;
  int count = 0;

  if (flags & UL_INSEN)
  {
    retval = regcomp(&reg, pattern, REG_EXTENDED|REG_ICASE|REG_NOSUB);
  }
  else
  {
    retval = regcomp(&reg, pattern, REG_EXTENDED|REG_NOSUB);
  }

  if (retval != 0)
  {
    char errbuf[BUFSIZE];
    regerror(retval, &reg, errbuf, BUFSIZE);
    sendto_one(source_p, ":%s NOTICE %s :NLIST regular expression error: %s",
               me.name, source_p->name, errbuf);
    regfree(&reg);
    return;
  }

  DLINK_FOREACH(dl, lclient_list.head)
  {
    target_p = dl->data;

    if ((lim > 0) && (count == lim))
    {
      sendto_one(source_p, ":%s NOTICE %s :NLIST limit reached, %d clients "
                 "shown", me.name, source_p->name, count);
      return;
    }
    /* Using no flags here */
    if (regexec(&reg, target_p->name, 0, NULL, 0) == 0)
    {
      count++;
      sendto_one(source_p, ":%s NOTICE %s :User %d: %s (%s@%s) [%s]",
                 me.name, source_p->name, count, target_p->name,
                 target_p->username, target_p->host, target_p->user->server);
    }
  }
  sendto_one(source_p, ":%s NOTICE %s :End of NLIST.", me.name, source_p->name);
  regfree(&reg);
}

static void nl_global_match(struct Client *client_p, struct Client *source_p,
                            char *pattern, int lim)
{
  dlink_node *dl;
  struct Client *target_p;
  int count = 0;

  report_operspy(source_p, "NLIST [GLOBAL][MATCH]", pattern);

  DLINK_FOREACH(dl, global_client_list.head)
  { 
    target_p = dl->data;

    if (!IsPerson(target_p))
      continue;

    if ((lim > 0) && (count == lim))
    {
      sendto_one(source_p, ":%s NOTICE %s :NLIST limit reached, %d clients "
                 "shown", me.name, source_p->name, count);
      return;
    }
    if (match(pattern, target_p->name))
    {
      count++;
      sendto_one(source_p, ":%s NOTICE %s :User %d: %s (%s@%s) [%s]",
                 me.name, source_p->name, count, target_p->name,
                 target_p->username, target_p->host, target_p->user->server);
    }
  }
  sendto_one(source_p, ":%s NOTICE %s :End of NLIST.", me.name, source_p->name);
}

static void nl_global_regex(struct Client *client_p, struct Client *source_p,
                            char *pattern, int lim, unsigned int flags)
{
  int retval;
  dlink_node *dl;
  struct Client *target_p;
  regex_t reg;
  int count = 0;
  char osbuffer[BUFSIZE];

  if (flags & UL_INSEN)
  {
    retval = regcomp(&reg, pattern, REG_EXTENDED|REG_ICASE|REG_NOSUB);
  }
  else
  {
    retval = regcomp(&reg, pattern, REG_EXTENDED|REG_NOSUB);
  }

  if (retval != 0)
  {
    char errbuf[BUFSIZE];
    regerror(retval, &reg, errbuf, BUFSIZE);
    sendto_one(source_p, ":%s NOTICE %s :NLIST regular expression error: %s",
               me.name, source_p->name, errbuf);
    regfree(&reg);
    return;
  }

  snprintf(osbuffer, BUFSIZE, "[GLOBAL][REGEX]%s %s",
           (flags & UL_INSEN) ? "[INSENS]" : "[SENS]", pattern);
  report_operspy(source_p, "NLIST", osbuffer);

  DLINK_FOREACH(dl, global_client_list.head)
  {
    target_p = dl->data;

    if (!IsPerson(target_p))
      continue;

    if ((lim > 0) && (count == lim))
    {
      sendto_one(source_p, ":%s NOTICE %s :NLIST limit reached, %d clients "
                 "shown", me.name, source_p->name, count);
      return;
    }
    /* Using no flags here */
    if (regexec(&reg, target_p->name, 0, NULL, 0) == 0)
    {
      count++;
      sendto_one(source_p, ":%s NOTICE %s :User %d: %s (%s@%s) [%s]",
                 me.name, source_p->name, count, target_p->name,
                 target_p->username, target_p->host, target_p->user->server);
    }
  }
  sendto_one(source_p, ":%s NOTICE %s :End of NLIST.", me.name, source_p->name);
  regfree(&reg);
}

static void nl_usage(struct Client *source_p)
{
  sendto_one(source_p, ":%s NOTICE %s :Syntax:  /NLIST [-l|-g] [-m|-r] [-i|-s] "
                       "[-x limit] pattern", me.name, source_p->name);
}

/*
** mo_glist
** Syntax:  NLIST [-l|-g] [-m|-r] [-i|-s] [-x limit] user@host
** Defaults:  local scope, limit to 0, match() variant, case insensitive (re)
*/
static int mo_glist(struct Client *client_p, struct Client *source_p,
                    int parc, char *parv[])
{		 
  int lim = 0;
  char *pattern;
  unsigned int flags = UL_LOCAL|UL_MATCH|UL_INSEN;
  char c;

  opterr = 0; /* Disable error reporting (since it goes to STDERR) */
  optind = 1; /* Always restore the index to the start of the parameters */
  while ( (c = getopt(parc, parv, "lgmrisx:")) != -1)
  {
    switch (c)
    {
      case 'l':
        flags &= ~UL_GLOBAL;
        flags |= UL_LOCAL;
        break;
      case 'g':
        flags &= ~UL_LOCAL;
        flags |= UL_GLOBAL;
        break;
      case 'm':
        flags &= ~UL_REGEX;
        flags |= UL_MATCH;
        break;
      case 'r':
        flags &= ~UL_MATCH;
        flags |= UL_REGEX;
        break;
      case 'i':
        flags &= ~UL_SENS;
        flags |= UL_INSEN;
        break;
      case 's':
        flags &= ~UL_INSEN;
        flags |= UL_SENS;
        break;
      case 'x':
        lim = atoi(optarg);
        break;
      case '?': /* Returned on a syntax error from getopt() */
      default:
        nl_usage(source_p);
        return 0;
        break;
    }
  }

  /* This extracts the pointer to the pattern, which is placed after all
  ** all of the arguments to the ULIST command.
  */
  parc -= optind;
  parv += optind;

  if (parc < 1)
  {
    gl_usage(source_p);
    return 0;
  }

  pattern = parv[0];

  /* Send the information to the proper handlers */
  if (flags & UL_LOCAL)
  {
    if (flags & UL_MATCH)
    {
      gl_local_match(client_p, source_p, pattern, lim);
    }
    else /* UL_REGEX */
    {
      gl_local_regex(client_p, source_p, pattern, lim, flags);
    }
  }
  else /* UL_GLOBAL */
  {
    if (flags & UL_MATCH)
    {
      gl_global_match(client_p, source_p, pattern, lim);
    }
    else /* UL_REGEX */
    {
      gl_global_regex(client_p, source_p, pattern, lim, flags);
    }
  }

  return 0;
}

static void gl_local_match(struct Client *client_p, struct Client *source_p,
                           char *pattern, int lim)
{
  dlink_node *dl;
  struct Client *target_p;
  int count = 0;

  DLINK_FOREACH(dl, lclient_list.head)
  {
    target_p = dl->data;

    if ((lim > 0) && (count == lim))
    {
      sendto_one(source_p, ":%s NOTICE %s :GLIST limit reached, %d clients "
                 "shown", me.name, source_p->name, count);
      return;
    }
    if (match(pattern, target_p->info))
    {
      count++;
      sendto_one(source_p, ":%s NOTICE %s :User %d: %s (%s@%s) [%s]",
                 me.name, source_p->name, count, target_p->name,
                 target_p->username, target_p->host, target_p->user->server);
    }
  }
  sendto_one(source_p, ":%s NOTICE %s :End of GLIST.", me.name, source_p->name);
}

static void gl_local_regex(struct Client *client_p, struct Client *source_p,
                           char *pattern, int lim, unsigned int flags)
{
  int retval;
  dlink_node *dl;
  struct Client *target_p;
  regex_t reg;
  int count = 0;

  if (flags & UL_INSEN)
  {
    retval = regcomp(&reg, pattern, REG_EXTENDED|REG_ICASE|REG_NOSUB);
  }
  else
  {
    retval = regcomp(&reg, pattern, REG_EXTENDED|REG_NOSUB);
  }

  if (retval != 0)
  {
    char errbuf[BUFSIZE];
    regerror(retval, &reg, errbuf, BUFSIZE);
    sendto_one(source_p, ":%s NOTICE %s :GLIST regular expression error: %s",
               me.name, source_p->name, errbuf);
    regfree(&reg);
    return;
  }

  DLINK_FOREACH(dl, lclient_list.head)
  {
    target_p = dl->data;

    if ((lim > 0) && (count == lim))
    {
      sendto_one(source_p, ":%s NOTICE %s :GLIST limit reached, %d clients "
                 "shown", me.name, source_p->name, count);
      return;
    }
    /* Using no flags here */
    if (regexec(&reg, target_p->info, 0, NULL, 0) == 0)
    {
      count++;
      sendto_one(source_p, ":%s NOTICE %s :User %d: %s (%s@%s) [%s]",
                 me.name, source_p->name, count, target_p->name,
                 target_p->username, target_p->host, target_p->user->server);
    }
  }
  sendto_one(source_p, ":%s NOTICE %s :End of GLIST.", me.name, source_p->name);
  regfree(&reg);
}

static void gl_global_match(struct Client *client_p, struct Client *source_p,
                            char *pattern, int lim)
{
  dlink_node *dl;
  struct Client *target_p;
  int count = 0;

  report_operspy(source_p, "GLIST [GLOBAL][MATCH]", pattern);

  DLINK_FOREACH(dl, global_client_list.head)
  { 
    target_p = dl->data;

    if (!IsPerson(target_p))
      continue;

    if ((lim > 0) && (count == lim))
    {
      sendto_one(source_p, ":%s NOTICE %s :GLIST limit reached, %d clients "
                 "shown", me.name, source_p->name, count);
      return;
    }
    if (match(pattern, target_p->info))
    {
      count++;
      sendto_one(source_p, ":%s NOTICE %s :User %d: %s (%s@%s) [%s]",
                 me.name, source_p->name, count, target_p->name,
                 target_p->username, target_p->host, target_p->user->server);
    }
  }
  sendto_one(source_p, ":%s NOTICE %s :End of GLIST.", me.name, source_p->name);
}

static void gl_global_regex(struct Client *client_p, struct Client *source_p,
                            char *pattern, int lim, unsigned int flags)
{
  int retval;
  dlink_node *dl;
  struct Client *target_p;
  regex_t reg;
  int count = 0;
  char osbuffer[BUFSIZE];

  if (flags & UL_INSEN)
  {
    retval = regcomp(&reg, pattern, REG_EXTENDED|REG_ICASE|REG_NOSUB);
  }
  else
  {
    retval = regcomp(&reg, pattern, REG_EXTENDED|REG_NOSUB);
  }

  if (retval != 0)
  {
    char errbuf[BUFSIZE];
    regerror(retval, &reg, errbuf, BUFSIZE);
    sendto_one(source_p, ":%s NOTICE %s :GLIST regular expression error: %s",
               me.name, source_p->name, errbuf);
    regfree(&reg);
    return;
  }

  snprintf(osbuffer, sizeof(osbuffer), "[GLOBAL][REGEX]%s %s",
           (flags & UL_INSEN) ? "[INSENS]" : "[SENS]", pattern);
  report_operspy(source_p, "GLIST", osbuffer);

  DLINK_FOREACH(dl, global_client_list.head)
  {
    target_p = dl->data;

    if (!IsPerson(target_p))
      continue;

    if ((lim > 0) && (count == lim))
    {
      sendto_one(source_p, ":%s NOTICE %s :GLIST limit reached, %d clients "
                 "shown", me.name, source_p->name, count);
      return;
    }
    /* Using no flags here */
    if (regexec(&reg, target_p->info, 0, NULL, 0) == 0)
    {
      count++;
      sendto_one(source_p, ":%s NOTICE %s :User %d: %s (%s@%s) [%s]",
                 me.name, source_p->name, count, target_p->name,
                 target_p->username, target_p->host, target_p->user->server);
    }
  }
  sendto_one(source_p, ":%s NOTICE %s :End of GLIST.", me.name, source_p->name);
  regfree(&reg);
}

static void gl_usage(struct Client *source_p)
{
  sendto_one(source_p, ":%s NOTICE %s :Syntax:  /GLIST [-l|-g] [-m|-r] [-i|-s] "
                       "[-x limit] pattern", me.name, source_p->name);
}
